@@ -32,9 +32,6 @@
 #include <libart_lgpl/art_rgba.h>
 #include <libart_lgpl/art_affine.h>
 
-/* Our default font */
-#define DEFAULT_FONT "Times Roman"
-
 gdouble
 rsvg_viewport_percentage (gdouble width, gdouble height)
 {
@@ -58,14 +55,14 @@ rsvg_state_init (RsvgState *state)
 	state->join = ART_PATH_STROKE_JOIN_MITER;
 	state->stop_opacity = 0xff;
 	
-	state->font_family  = g_strdup (DEFAULT_FONT);
+	state->font_family  = g_strdup ("Times New Roman");
 	state->font_size    = 12.0;
 	state->font_style   = PANGO_STYLE_NORMAL;
 	state->font_variant = PANGO_VARIANT_NORMAL;
 	state->font_weight  = PANGO_WEIGHT_NORMAL;
 	state->font_stretch = PANGO_STRETCH_NORMAL;
-
-	state->visible = TRUE;
+	state->text_dir     = PANGO_DIRECTION_LTR;
+	state->visible      = TRUE;
 }
 
 void
@@ -200,18 +197,33 @@ rsvg_parse_style_arg (RsvgHandle *ctx, RsvgState *state, const char *str)
 		}
 	else if (rsvg_css_param_match (str, "text-decoration"))
 		{
-			if (!strcmp (str, "inherit"))
+			if (!strcmp (str + arg_off, "inherit"))
 				state->font_decor = parent_state->font_decor;
 			else 
 				{
-					if (strstr (str, "underline"))
+					if (strstr (str + arg_off, "underline"))
 						state->font_decor |= TEXT_UNDERLINE;
-					if (strstr (str, "overline"))
+					if (strstr (str + arg_off, "overline"))
 						state->font_decor |= TEXT_OVERLINE;
-					if (strstr (str, "strike") || strstr (str, "line-through")) /* strike though or line-through */
+					if (strstr (str + arg_off, "strike") || strstr (str + arg_off, "line-through")) /* strike though or line-through */
 						state->font_decor |= TEXT_STRIKE;
 				}
 		}
+	else if (rsvg_css_param_match (str, "writing-mode"))
+		{
+			/* lr-tb | rl-tb | tb-rl | lr | rl | tb | inherit */
+			if (!strcmp (str + arg_off, "inherit"))
+				state->text_dir = parent_state->text_dir;
+			else if (!strcmp (str + arg_off, "rl"))
+				state->text_dir = PANGO_DIRECTION_RTL;
+			else if (!strcmp (str + arg_off, "tb-rl") || 
+					 !strcmp (str + arg_off, "rl-tb")) /* not sure of tb-rl vs. rl-tb */
+				state->text_dir = PANGO_DIRECTION_TTB_RTL;
+			else if (!strcmp (str + arg_off, "lr-tb"))
+				state->text_dir = PANGO_DIRECTION_TTB_LTR;
+			else
+				state->text_dir = PANGO_DIRECTION_LTR;
+		}
 	else if (rsvg_css_param_match (str, "stop-color"))
 		{
 			state->stop_color = rsvg_css_parse_color (str + arg_off);
@@ -309,6 +321,7 @@ rsvg_is_style_arg(const char *str)
 			g_hash_table_insert (styles, "stroke-width",      GINT_TO_POINTER (TRUE));
 			g_hash_table_insert (styles, "text-decoration",   GINT_TO_POINTER (TRUE));
 			g_hash_table_insert (styles, "visibility",        GINT_TO_POINTER (TRUE));
+			g_hash_table_insert (styles, "writing-mode",      GINT_TO_POINTER (TRUE));
 		}
 	
 	/* this will default to 0 (FALSE) on a failed lookup */
@@ -352,22 +365,184 @@ rsvg_parse_style (RsvgHandle *ctx, RsvgState *state, const char *str)
 		}
 }
 
-/*
- * Extremely poor man's CSS parser. Not robust. Not compliant.
- * Should work well enough for our needs ;-)
- * See also: http://www.w3.org/TR/REC-CSS2/syndata.html
- * I should use that sometime in order to make a complaint parser
- */
-void
-rsvg_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen)
+static void
+rsvg_css_define_style (RsvgHandle *ctx, const char * style_name, const char * style_def)
 {
+	GString * str = g_string_new (style_def);
+	char * existing = NULL;
+
+	/* push name/style pair into HT */
+	existing = (char *)g_hash_table_lookup (ctx->css_props, style_name);
+	if (existing != NULL)
+		g_string_append_len (str, existing, strlen (existing));
+	
+	/* will destroy the existing key and value for us */
+	g_hash_table_insert (ctx->css_props, (gpointer)g_strdup (style_name), (gpointer)str->str);
+	g_string_free (str, FALSE);
+}
+
+#ifdef HAVE_LIBCROCO
+
+#include <libcroco.h>
+
+typedef struct _CSSUserData
+{
+	RsvgHandle *ctx;
+	GString    *def;
+} CSSUserData;
+
+static void
+css_user_data_init (CSSUserData *user_data, RsvgHandle * ctx)
+{
+	user_data->ctx = ctx;
+	user_data->def = NULL;
+}
+
+static void
+ccss_start_selector (CRDocHandler *a_handler,
+					 CRSelector *a_selector_list)
+{
+	CSSUserData * user_data;
+
+	g_return_if_fail (a_handler);
+
+	user_data = (CSSUserData *)a_handler->app_data;
+	user_data->def = g_string_new (NULL);
+}
+
+static void
+ccss_end_selector (CRDocHandler *a_handler,
+				   CRSelector *a_selector_list)
+{
+	CSSUserData * user_data;
+	CRSelector  * cur;
+
+	g_return_if_fail (a_handler);
+
+	user_data = (CSSUserData *)a_handler->app_data;
+
+	if (a_selector_list)
+		for (cur = a_selector_list; cur; cur = cur->next) {
+			if (cur->simple_sel) {
+			   guchar * style_name = cr_simple_sel_to_string (cur->simple_sel);
+			   if (style_name) {
+				   rsvg_css_define_style (user_data->ctx, style_name, user_data->def->str);
+				   g_free (style_name);
+			   }
+			}
+		}
+
+	g_string_free (user_data->def, TRUE);
+}
+
+static void
+ccss_property (CRDocHandler *a_handler, GString *a_name, CRTerm *a_expr)
+{
+	CSSUserData * user_data;
+	char * expr = NULL;
+
+	g_return_if_fail (a_handler);
+
+	user_data = (CSSUserData *)a_handler->app_data;
+
+	if (a_name && a_name->str && a_expr)
+		{
+			g_string_append_len (user_data->def, a_name->str, a_name->len);
+			g_string_append (user_data->def, ": ");
+			expr = cr_term_to_string (a_expr);
+			g_string_append_len (user_data->def, expr, strlen (expr));
+			g_free (expr);
+			g_string_append (user_data->def, "; ");
+        }
+}
+
+static void
+ccss_error (CRDocHandler *a_handler)
+{
+	/* yup, like i care about CSS parsing errors ;-)
+	   ignore, chug along */
+	g_warning ("CSS parsing error\n");
+}
+
+static void
+ccss_unrecoverable_error (CRDocHandler *a_handler)
+{
+	/* yup, like i care about CSS parsing errors ;-)
+	   ignore, chug along */
+	g_warning ("CSS unrecoverable error\n");
+}
+
+static void
+init_sac_handler (CRDocHandler *a_handler)
+{
+	a_handler->start_document        = NULL;
+	a_handler->end_document          = NULL;
+	a_handler->import_style          = NULL;
+	a_handler->namespace_declaration = NULL;
+	a_handler->comment               = NULL;
+	a_handler->start_selector        = ccss_start_selector;
+	a_handler->end_selector          = ccss_end_selector;
+	a_handler->property              = ccss_property;
+	a_handler->start_font_face       = NULL;
+	a_handler->end_font_face         = NULL;
+	a_handler->start_media           = NULL;
+	a_handler->end_media             = NULL;
+	a_handler->start_page            = NULL;
+	a_handler->end_page              = NULL;
+	a_handler->ignorable_at_rule     = NULL;
+	a_handler->error                 = ccss_error;
+	a_handler->unrecoverable_error   = ccss_unrecoverable_error;
+}
+
+static void
+rsvg_real_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen)
+{
+	enum CRStatus status = CR_OK;
+	CRParser *parser = NULL;
+	CRDocHandler * css_handler = NULL;
+	CSSUserData user_data;
+
+	css_handler = cr_doc_handler_new ();
+	init_sac_handler (css_handler);
+
+	css_user_data_init (&user_data, ctx);
+	css_handler->app_data = &user_data;
+
+	/* TODO: fix libcroco to take in const strings */
+	parser = cr_parser_new_from_buf ((guchar *)buff, (gulong)buflen, CR_UTF_8, FALSE);	
+	status = cr_parser_set_sac_handler (parser, css_handler);
+    
+	if (status != CR_OK)
+        {
+			g_warning ("Error setting CSS SAC handler\n");
+			cr_parser_destroy (parser);			
+			return;
+        }        
+	
+	status = cr_parser_set_use_core_grammar (parser, FALSE);
+	status = cr_parser_parse (parser);
+	
+	cr_parser_destroy (parser);
+}
+
+#else /* !HAVE_LIBCROCO */
+
+#warning Building without libcroco support. Will experience sub-optimal CSS parsing.
+
+static void
+rsvg_real_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen)
+{
+	/*
+	 * Extremely poor man's CSS parser. Not robust. Not compliant.
+	 * See also: http://www.w3.org/TR/REC-CSS2/syndata.html
+	 */
+
 	size_t loc = 0;
 	
 	while (loc < buflen)
 		{
 			GString * style_name  = g_string_new (NULL);
 			GString * style_props = g_string_new (NULL);
-			const char * existing    = NULL;
 
 			/* advance to the style's name */
 			while (loc < buflen && g_ascii_isspace (buff[loc]))
@@ -402,16 +577,9 @@ rsvg_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen)
 						}
 				}
 
-			/* push name/style pair into HT */
-			existing = (const char *)g_hash_table_lookup (ctx->css_props, style_name->str);
-			if (existing != NULL)
-				g_string_append_len (style_props, existing, strlen (existing));
-
-			/* will destroy the existing key and value for us */
-			g_hash_table_insert (ctx->css_props, style_name->str, style_props->str);
-
-			g_string_free (style_name, FALSE);
-			g_string_free (style_props, FALSE);
+			rsvg_css_define_style (ctx, style_name->str, style_props->str);
+			g_string_free (style_name, TRUE);
+			g_string_free (style_props, TRUE);
 			
 			loc++;
 			while (loc < buflen && g_ascii_isspace (buff[loc]))
@@ -419,6 +587,15 @@ rsvg_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen)
 		}
 }
 
+#endif /* HAVE_LIBCROCO */
+
+void
+rsvg_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen)
+{
+	/* delegate off to the builtin or libcroco implementation */
+	rsvg_real_parse_cssbuffer (ctx, buff, buflen);
+}
+
 /* Parse an SVG transform string into an affine matrix. Reference: SVG
    working draft dated 1999-07-06, section 8.5. Return TRUE on
    success. */
@@ -563,13 +740,7 @@ rsvg_parse_transform_attr (RsvgHandle *ctx, RsvgState *state, const char *str)
 	double affine[6];
 	
 	if (rsvg_parse_transform (affine, str))
-		{
-			art_affine_multiply (state->affine, affine, state->affine);
-		}
-	else
-		{
-			/* parse error for transform attribute. todo: report */
-		}
+		art_affine_multiply (state->affine, affine, state->affine);
 }
 
 static gboolean
@@ -589,6 +760,7 @@ rsvg_lookup_apply_css_style (RsvgHandle *ctx, const char * target)
 /**
  * rsvg_parse_style_attrs: Parse style attribute.
  * @ctx: Rsvg context.
+ * @state: Rsvg state
  * @tag: The SVG tag we're processing (eg: circle, ellipse), optionally %NULL
  * @klazz: The space delimited class list, optionally %NULL
  * @atts: Attributes in SAX style.
@@ -597,7 +769,8 @@ rsvg_lookup_apply_css_style (RsvgHandle *ctx, const char * target)
  * stack.
  **/
 void
-rsvg_parse_style_attrs (RsvgHandle *ctx, 
+rsvg_parse_style_attrs (RsvgHandle * ctx, 
+						RsvgState  * state,
 						const char * tag,
 						const char * klazz,
 						const char * id,
@@ -608,9 +781,21 @@ rsvg_parse_style_attrs (RsvgHandle *ctx,
 	gboolean found = FALSE;
 	GString * klazz_list = NULL;
 	
-	/* handle the all-encompassing "star" entry first, ignoring found-ness */
+	/* Try to properly support all of the following, including inheritance:
+	 * *
+	 * #id
+	 * tag
+	 * tag#id
+	 * tag.class
+	 * tag.class#id
+	 *
+	 * This is basically a semi-compliant CSS2 selection engine
+	 */
+
+	/* * */
 	rsvg_lookup_apply_css_style (ctx, "*");
 
+	/* #id */
 	if (id != NULL)
 		{
 			target = g_strdup_printf ("#%s", id);
@@ -618,52 +803,66 @@ rsvg_parse_style_attrs (RsvgHandle *ctx,
 			g_free (target);
 		}
 
-	/* todo: see if klazz or tag should accumulate onto tag.klazz */
+	/* tag */
+	if (tag != NULL)
+		rsvg_lookup_apply_css_style (ctx, tag);
 
-	if (tag != NULL && klazz != NULL)
+	/* tag#id */
+	if (tag != NULL && id != NULL)
 		{
-			target = g_strdup_printf ("%s.%s", tag, klazz);
-			found = rsvg_lookup_apply_css_style (ctx, target);
+			target = g_strdup_printf ("%s#%s", tag, id);
+			rsvg_lookup_apply_css_style (ctx, target);
 			g_free (target);
 		}
 	
-	if (found == FALSE)
+	if (klazz != NULL)
 		{
-			if (tag != NULL)
-				rsvg_lookup_apply_css_style (ctx, tag);
-			
-			if (klazz != NULL)
+			i = strlen (klazz);
+			while (j < i)
 				{
-					i = strlen (klazz);
-					while (j < i)
+					found = FALSE;
+					klazz_list = g_string_new (".");
+					
+					while (j < i && g_ascii_isspace(klazz[j]))
+						j++;
+					
+					while (j < i && !g_ascii_isspace(klazz[j]))
+						g_string_append_c (klazz_list, klazz[j++]);
+					
+					/* tag.class */
+					if (tag != NULL && klazz_list->len != 1)
 						{
-							klazz_list = g_string_new (".");
-							
-							while (j < i && g_ascii_isspace(klazz[j]))
-								j++;
-							
-							while (j < i && !g_ascii_isspace(klazz[j]))
-								g_string_append_c (klazz_list, klazz[j++]);
-							
-							rsvg_lookup_apply_css_style (ctx, klazz_list->str);
-							g_string_free (klazz_list, TRUE);
+							target = g_strdup_printf ("%s%s", tag, klazz_list->str);
+							found = found || rsvg_lookup_apply_css_style (ctx, target);
+							g_free (target);
 						}
+					
+					/* tag.class#id */
+					if (tag != NULL && klazz_list->len != 1 && id != NULL)
+						{
+							target = g_strdup_printf ("%s%s#%s", tag, klazz_list->str, id);
+							found = found || rsvg_lookup_apply_css_style (ctx, target);
+							g_free (target);
+						}
+					
+					/* didn't find anything more specific, just apply the class style */
+					if (!found) {
+						found = found || rsvg_lookup_apply_css_style (ctx, klazz_list->str);
+					}
+					g_string_free (klazz_list, TRUE);
 				}
 		}
-	
+
 	if (atts != NULL)
 		{
 			for (i = 0; atts[i] != NULL; i += 2)
 				{
 					if (!strcmp ((char *)atts[i], "style"))
-						rsvg_parse_style (ctx, &ctx->state[ctx->n_state - 1],
-										  (char *)atts[i + 1]);
+						rsvg_parse_style (ctx, state, (char *)atts[i + 1]);
 					else if (!strcmp ((char *)atts[i], "transform"))
-						rsvg_parse_transform_attr (ctx, &ctx->state[ctx->n_state - 1],
-												   (char *)atts[i + 1]);
+						rsvg_parse_transform_attr (ctx, state, (char *)atts[i + 1]);
 					else if (rsvg_is_style_arg ((char *)atts[i]))
-						rsvg_parse_style_pair (ctx, &ctx->state[ctx->n_state - 1],
-											   (char *)atts[i], (char *)atts[i + 1]);
+						rsvg_parse_style_pair (ctx, state, (char *)atts[i], (char *)atts[i + 1]);
 				}
 		}
 }
@@ -790,3 +989,11 @@ rsvg_pop_opacity_group (RsvgHandle *ctx, int opacity)
 	g_object_unref (tos);
 	ctx->pixbuf = nos;
 }
+
+RsvgState * 
+rsvg_state_current (RsvgHandle *ctx)
+{
+	if (ctx->n_state > 0)
+		return &ctx->state[ctx->n_state - 1];
+	return NULL;
+}
