@@ -32,10 +32,10 @@
 #include <libart_lgpl/art_affine.h>
 #include <libart_lgpl/art_render_mask.h>
 
-#if ENABLE_TEXT_DECOR
-
 #include "rsvg-shapes.h"
 
+#if 0
+
 static void 
 rsvg_draw_hline (RsvgHandle *ctx, double x, double w, double y)
 {
@@ -57,7 +57,7 @@ rsvg_draw_hline (RsvgHandle *ctx, double x, double w, double y)
 	g_string_free (d, TRUE);
 }
 
-#endif /* ENABLE_TEXT_DECOR */
+#endif
 
 static char *
 make_valid_utf8 (const char *str)
@@ -124,7 +124,7 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	PangoRectangle ink_rect, line_ink_rect;
 	FT_Bitmap bitmap;
 	
-	state = &ctx->state[ctx->n_state - 1];
+	state = rsvg_state_current (ctx);
 	if (state->fill == NULL && state->font_size <= 0)
 		return;
 
@@ -146,7 +146,7 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	
 	for (end = len; end > beg; end--)
 		if (!g_ascii_isspace (ch[end - 1]))
-      break;
+			break;
 	
 	if (end - beg == 0)
 		{
@@ -169,23 +169,10 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 		}
 	
 	if (ctx->pango_context == NULL)
-		ctx->pango_context = pango_ft2_get_context ((guint)ctx->dpi, (guint)ctx->dpi);
-	
-	has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
-	
-	render = art_render_new (0, 0,
-							 gdk_pixbuf_get_width (pixbuf),
-							 gdk_pixbuf_get_height (pixbuf),
-							 gdk_pixbuf_get_pixels (pixbuf),
-							 gdk_pixbuf_get_rowstride (pixbuf),
-							 gdk_pixbuf_get_n_channels (pixbuf) -
-							 (has_alpha ? 1 : 0),
-							 gdk_pixbuf_get_bits_per_sample (pixbuf),
-							 has_alpha ? ART_ALPHA_SEPARATE : ART_ALPHA_NONE,
-							 NULL);
-	
+		ctx->pango_context = pango_ft2_get_context ((guint)ctx->dpi, (guint)ctx->dpi);   
+		
 	layout = pango_layout_new (ctx->pango_context);
-	pango_layout_set_text (layout, string, end - beg);
+	pango_layout_set_text (layout, string, -1);
 	font = pango_font_description_copy (pango_context_get_font_description (ctx->pango_context));
 
 	/* we need to resize the font by our X or Y scale (ideally could stretch in both directions...)
@@ -199,7 +186,10 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	pango_font_description_set_variant (font, state->font_variant);
 	pango_font_description_set_weight (font, state->font_weight);
 	pango_font_description_set_stretch (font, state->font_stretch);
-  
+	pango_layout_set_alignment (layout, (state->text_dir == PANGO_DIRECTION_LTR || 
+										 state->text_dir == PANGO_DIRECTION_TTB_LTR) ? 
+								PANGO_ALIGN_LEFT : PANGO_ALIGN_RIGHT);
+
 	pango_layout_set_font_description (layout, font);
 	pango_font_description_free (font);
 	
@@ -211,16 +201,27 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	else
 		pango_layout_line_get_pixel_extents (line, &line_ink_rect, NULL);
 	
-	bitmap.rows = ink_rect.height;
-	bitmap.width = ink_rect.width;
-	bitmap.pitch = (bitmap.width + 3) & ~3;
-	bitmap.buffer = g_malloc0 (bitmap.rows * bitmap.pitch);
-	bitmap.num_grays = 0x100;
+	bitmap.rows       = ink_rect.height;
+	bitmap.width      = ink_rect.width;
+	bitmap.pitch      = bitmap.width;
+	bitmap.buffer     = g_malloc0 (bitmap.rows * bitmap.pitch);
+	bitmap.num_grays  = 256;
 	bitmap.pixel_mode = ft_pixel_mode_grays;
 	
 	pango_ft2_render_layout (&bitmap, layout, -ink_rect.x, -ink_rect.y);
-	
 	g_object_unref (layout);
+
+	has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
+	render = art_render_new (0, 0,
+							 gdk_pixbuf_get_width (pixbuf),
+							 gdk_pixbuf_get_height (pixbuf),
+							 gdk_pixbuf_get_pixels (pixbuf),
+							 gdk_pixbuf_get_rowstride (pixbuf),
+							 gdk_pixbuf_get_n_channels (pixbuf) -
+							 (has_alpha ? 1 : 0),
+							 gdk_pixbuf_get_bits_per_sample (pixbuf),
+							 has_alpha ? ART_ALPHA_SEPARATE : ART_ALPHA_NONE,
+							 NULL);
 	
 	rsvg_render_paint_server (render, state->fill, NULL); /* todo: paint server ctx */
 	opacity = state->fill_opacity * state->opacity;
@@ -237,14 +238,15 @@ rsvg_text_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	
 	g_free (bitmap.buffer);
 	g_free (string);
-	
-#if ENABLE_TEXT_DECOR
+
+#if 0	
+	/* TODO: store the various line widths for these */
 	if (state->font_decor & TEXT_OVERLINE)
-		rsvg_draw_hline (ctx, 0, line_ink_rect.width, line_ink_rect.y - line_ink_rect.height);
+		rsvg_draw_hline (ctx, state->text_offset, line_ink_rect.width, -line_ink_rect.y + line_ink_rect.height);
 	if (state->font_decor & TEXT_UNDERLINE)
-		rsvg_draw_hline (ctx, 0, line_ink_rect.width, line_ink_rect.y);
+		rsvg_draw_hline (ctx, state->text_offset, line_ink_rect.width, -line_ink_rect.y);
 	if (state->font_decor & TEXT_STRIKE)
-		rsvg_draw_hline (ctx, 0, ink_rect.width, line_ink_rect.y - (line_ink_rect.height/2));
+		rsvg_draw_hline (ctx, state->text_offset, ink_rect.width, -line_ink_rect.y - (line_ink_rect.height/2));
 #endif
 	
 	state->text_offset += line_ink_rect.width;
@@ -260,37 +262,37 @@ rsvg_start_tspan (RsvgHandle *ctx, const xmlChar **atts)
 	const char * klazz = NULL, * id = NULL;
 	x = y = dx = dy = 0.;
 	
-  state = &ctx->state[ctx->n_state - 1];
-  
-  if (atts != NULL)
-	  {
-		  for (i = 0; atts[i] != NULL; i += 2)
-			  {
-				  if (!strcmp ((char *)atts[i], "x"))
-					  x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
-				  else if (!strcmp ((char *)atts[i], "y"))
-					  y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
-				  else if (!strcmp ((char *)atts[i], "dx"))
-					  dx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
-				  else if (!strcmp ((char *)atts[i], "dy"))
-					  dy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
-				  else if (!strcmp ((char *)atts[i], "class"))
-					  klazz = (const char *)atts[i + 1];
-				  else if (!strcmp ((char *)atts[i], "id"))
-					  id = (const char *)atts[i + 1];
-			  }
-	  }
-  
-  /* todo: transform() is illegal here */
-  x += dx ;
-  y += dy ;
-  
-  if (x > 0 && y > 0)
-	  {
-		  art_affine_translate (affine, x, y);
-		  art_affine_multiply (state->affine, affine, state->affine);
-	  }
-  rsvg_parse_style_attrs (ctx, "tspan", klazz, id, atts);
+	state = rsvg_state_current (ctx);
+	
+	if (atts != NULL)
+		{
+			for (i = 0; atts[i] != NULL; i += 2)
+				{
+					if (!strcmp ((char *)atts[i], "x"))
+						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+					else if (!strcmp ((char *)atts[i], "y"))
+						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+					else if (!strcmp ((char *)atts[i], "dx"))
+						dx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
+					else if (!strcmp ((char *)atts[i], "dy"))
+						dy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
+					else if (!strcmp ((char *)atts[i], "class"))
+						klazz = (const char *)atts[i + 1];
+					else if (!strcmp ((char *)atts[i], "id"))
+						id = (const char *)atts[i + 1];
+				}
+		}
+	
+	/* todo: transform() is illegal here */
+	x += dx ;
+	y += dy ;
+	
+	if (x > 0 && y > 0)
+		{
+			art_affine_translate (affine, x, y);
+			art_affine_multiply (state->affine, affine, state->affine);
+		}
+	rsvg_parse_style_attrs (ctx, state, "tspan", klazz, id, atts);
 }
 
 static void
@@ -360,7 +362,7 @@ rsvg_start_text (RsvgHandle *ctx, const xmlChar **atts)
 	
 	x = y = dx = dy = 0.;
 	
-	state = &ctx->state[ctx->n_state - 1];
+	state = rsvg_state_current (ctx);
 	
 	if (atts != NULL)
 		{
@@ -387,7 +389,7 @@ rsvg_start_text (RsvgHandle *ctx, const xmlChar **atts)
 	art_affine_translate (affine, x, y);
 	art_affine_multiply (state->affine, affine, state->affine);
 	
-	rsvg_parse_style_attrs (ctx, "text", klazz, id, atts);
+	rsvg_parse_style_attrs (ctx, state, "text", klazz, id, atts);
 	
 	handler->parent = ctx->handler;
 	ctx->handler = &handler->super;
